
# SPDX-License-Identifier: BSD-2

# this file is autogenerated during the build

_type_map = {
    ("TPMS_ALGORITHM_DESCRIPTION", "alg"): "TPM2_ALG",
    ("TPMS_ALGORITHM_DESCRIPTION", "attributes"): "TPMA_ALGORITHM",
    ("TPMT_HA", "hashAlg"): "TPM2_ALG",
    ("TPMU_NAME", "handle"): "TPM2_HANDLE",
    ("TPMS_PCR_SELECTION", "hash"): "TPM2_ALG",
    ("TPMT_TK_CREATION", "tag"): "TPM2_ST",
    ("TPMT_TK_CREATION", "hierarchy"): "TPM2_RH",
    ("TPMT_TK_VERIFIED", "tag"): "TPM2_ST",
    ("TPMT_TK_VERIFIED", "hierarchy"): "TPM2_RH",
    ("TPMT_TK_AUTH", "tag"): "TPM2_ST",
    ("TPMT_TK_AUTH", "hierarchy"): "TPM2_RH",
    ("TPMT_TK_HASHCHECK", "tag"): "TPM2_ST",
    ("TPMT_TK_HASHCHECK", "hierarchy"): "TPM2_RH",
    ("TPMS_ALG_PROPERTY", "alg"): "TPM2_ALG",
    ("TPMS_ALG_PROPERTY", "algProperties"): "TPMA_ALGORITHM",
    ("TPMS_TAGGED_PROPERTY", "property"): "TPM2_PT",
    ("TPMS_TAGGED_PCR_SELECT", "tag"): "TPM2_PT_PCR",
    ("TPMS_TAGGED_POLICY", "handle"): "TPM2_HANDLE",
    ("TPMS_ACT_DATA", "handle"): "TPM2_HANDLE",
    ("TPMS_CAPABILITY_DATA", "capability"): "TPM2_CAP",
    ("TPMS_COMMAND_AUDIT_INFO", "digestAlg"): "TPM2_ALG",
    ("TPMS_ATTEST", "magic"): "TPM2_GENERATED",
    ("TPMS_ATTEST", "type"): "TPM2_ST",
    ("TPMS_AUTH_COMMAND", "sessionHandle"): "TPM2_HANDLE",
    ("TPMS_AUTH_COMMAND", "sessionAttributes"): "TPMA_SESSION",
    ("TPMS_AUTH_RESPONSE", "sessionAttributes"): "TPMA_SESSION",
    ("TPMU_SYM_KEY_BITS", "exclusiveOr"): "TPM2_ALG",
    ("TPMU_SYM_MODE", "aes"): "TPM2_ALG",
    ("TPMU_SYM_MODE", "sm4"): "TPM2_ALG",
    ("TPMU_SYM_MODE", "camellia"): "TPM2_ALG",
    ("TPMU_SYM_MODE", "sym"): "TPM2_ALG",
    ("TPMT_SYM_DEF", "algorithm"): "TPM2_ALG",
    ("TPMT_SYM_DEF_OBJECT", "algorithm"): "TPM2_ALG",
    ("TPMS_SCHEME_HASH", "hashAlg"): "TPM2_ALG",
    ("TPMS_SCHEME_ECDAA", "hashAlg"): "TPM2_ALG",
    ("TPMS_SCHEME_XOR", "hashAlg"): "TPM2_ALG",
    ("TPMS_SCHEME_XOR", "kdf"): "TPM2_ALG",
    ("TPMT_KEYEDHASH_SCHEME", "scheme"): "TPM2_ALG",
    ("TPMT_SIG_SCHEME", "scheme"): "TPM2_ALG",
    ("TPMT_KDF_SCHEME", "scheme"): "TPM2_ALG",
    ("TPMT_ASYM_SCHEME", "scheme"): "TPM2_ALG",
    ("TPMT_RSA_SCHEME", "scheme"): "TPM2_ALG",
    ("TPMT_RSA_DECRYPT", "scheme"): "TPM2_ALG",
    ("TPMT_ECC_SCHEME", "scheme"): "TPM2_ALG",
    ("TPMS_ALGORITHM_DETAIL_ECC", "curveID"): "TPM2_ECC_CURVE",
    ("TPMS_SIGNATURE_RSA", "hash"): "TPM2_ALG",
    ("TPMS_SIGNATURE_ECC", "hash"): "TPM2_ALG",
    ("TPMT_SIGNATURE", "sigAlg"): "TPM2_ALG",
    ("TPMS_ECC_PARMS", "curveID"): "TPM2_ECC_CURVE",
    ("TPMT_PUBLIC_PARMS", "type"): "TPM2_ALG",
    ("TPMT_PUBLIC", "type"): "TPM2_ALG",
    ("TPMT_PUBLIC", "nameAlg"): "TPM2_ALG",
    ("TPMT_PUBLIC", "objectAttributes"): "TPMA_OBJECT",
    ("TPMT_SENSITIVE", "sensitiveType"): "TPM2_ALG",
    ("TPMS_NV_PUBLIC", "nvIndex"): "TPM2_RH",
    ("TPMS_NV_PUBLIC", "nameAlg"): "TPM2_ALG",
    ("TPMS_NV_PUBLIC", "attributes"): "TPMA_NV",
    ("TPMS_CONTEXT", "savedHandle"): "TPM2_HANDLE",
    ("TPMS_CONTEXT", "hierarchy"): "TPM2_RH",
    ("TPMS_CREATION_DATA", "locality"): "TPMA_LOCALITY",
    ("TPMS_CREATION_DATA", "parentNameAlg"): "TPM2_ALG",

}

_element_type_map = {
    "TPML_CC": "TPM2_CC",
    "TPML_CCA": "TPMA_CC",
    "TPML_ALG": "TPM2_ALG",
    "TPML_HANDLE": "TPM2_HANDLE",
    "TPML_ECC_CURVE": "TPM2_ECC_CURVE",

}
